//************************************************************************************************************************** 
//* Blue HttpMail Extension
//* Copyright (c) 2003-2004 Josh Harler
//* 
//* Blue - General Purpose C++ Library
//* Copyright (c) 2002-2004 Josh Harler
//* 
//* This software is provided 'as-is', without any express or implied warranty. In no event
//* will the authors be held liable for any damages arising from the use of this software.
//* 
//* Permission is granted to anyone to use this software for any purpose, including commercial
//* applications, and to alter it and redistribute it freely, subject to the following restrictions:
//* 
//* 	1. The origin of this software must not be misrepresented; you must not claim that you
//* 	wrote the original software. If you use this software in a product, an acknowledgment in the
//* 	product documentation would be appreciated but is not required.
//* 
//* 	2. Altered source versions must be plainly marked as such, and must not be misrepresented as
//* 	being the original software.
//* 
//* 	3. This notice may not be removed or altered from any source distribution.
//*
//*
//* file   Blue/Extension/HttpMail/Mailbox.cpp
//**

// Private Headers =========================================================================================================

// matching header
#include "Mailbox.h"

// Extension headers
#include "Folder.h"

// Blue library headers
#include "Blue/Data/BufferDevice.h"
#include "Blue/Data/FilterByte.h"
#include "Blue/Data/File.h"
#include "Blue/Util/StringTokenizer.h"
#include "Blue/Util/DateTime.h"

// Blue Xml Extension headers
#include "Blue/Extension/Xml/DomParser.h"
#include "Blue/Extension/Xml/XPathExpression.h"


// Private Defines/Enums/Typedefs/Etc ======================================================================================

// Private Classes/Structs =================================================================================================

using namespace blue;
using namespace blue::common;
using namespace blue::util;
using namespace blue::ext;
using namespace blue::ext::http;

namespace {

	/****
	 * \class HttpMailPasswordRetriever
	 * \brief Used to give the Session the username and password.
	 */
	class HttpMailPasswordRetriever :public Session::PasswordRetriever
	{
	public:
		HttpMailPasswordRetriever( String user, String password )
			:m_user(user), m_password(password)
		{}

		virtual ~HttpMailPasswordRetriever()
		{}

		virtual bool getUserPassword( const Response& response, const Url& url, String& user, String& pass )
		{
			String urlStr = url.getUrl();
			if( m_lastUrl == urlStr ) {
				return (false);
			}

			user = m_user;
			pass = m_password;
			
			m_lastUrl = urlStr;
			return (true);
		}

	private:
		String m_user;
		String m_password;
		String m_lastUrl;
	};


	/****
	 * \class LogFilter
	 * \brief Used to log network communication to file for debugging purposes.
	 */
	class LogFilter : public data::FilterByte
	{
	public:
		LogFilter( String filename, String msg ) :m_file(filename)
		{
			m_file.open(data::File::WRITE);

			String message = "\r\n---------------------------------------------------------------\r\n"
							 + msg +
							 "\r\n---------------------------------------------------------------\r\n\r\n";
			m_file.write(message);
		}

		virtual ~LogFilter() {}

		virtual void processByte( uint8_t& byte )
		{
			m_file.writeUint8(byte);
			m_file.flush();
		}

	private:
		data::File m_file;
	};


	/****
	 * \class LogSession
	 * \brief Used to log network communication to file for debugging purposes.
	 */
	class LogSession: public Session
	{
	public:
		LogSession( String filename ) :m_requests(0), m_responses(0), m_filename(filename)
		{
			if( data::File(m_filename).exists() ) {
				data::File(m_filename).remove();
			}
		}

		virtual ~LogSession()
		{}

	protected:
		virtual void prepareInputStream( data::InputStream& input )
		{
			resetFilterChain(m_inputChain);

			String msg = "Response " + String(++m_responses);
			m_inputChain.addFilter( new LogFilter(m_filename, msg), true );
			input.setFilterChain(&m_inputChain);
		}

		virtual void prepareOutputStream( data::OutputStream& output )
		{
			resetFilterChain(m_outputChain);

			String msg = "Request " + String(++m_requests);
			m_outputChain.addFilter( new LogFilter(m_filename, msg), true );
			output.setFilterChain(&m_outputChain);
		}

		void resetFilterChain( data::FilterChain& chain )
		{
			int count = chain.getFilterCount();
			for( int i = 0; i < count; ++i ) {
				chain.removeFilter(i);
			}
		}

		String m_filename;
		int m_requests, m_responses;
		data::FilterChain m_inputChain;
		data::FilterChain m_outputChain;
	};

} // namespace


// Private Global Variables ================================================================================================

namespace {
	// The default user-agent field sent with requests
	const String g_defUserAgent = String("Outlook Express/5.5 (MSIE 5.0; Windows 98; TmstmpExt)", String::STATIC);

	// The default content that is posted with the login request
	const String g_defLoginContent =
		String("<?xml version=\"1.0\"?>\n"
               "<D:propfind xmlns:D=\"DAV:\" xmlns:h=\"http://schemas.microsoft.com/hotmail/\" xmlns:hm=\"urn:schemas:httpmail:\">\n"
               "	<D:prop>\n"
               "		<h:adbar/>\n"
               "		<hm:contacts/>\n"
               "		<hm:inbox/>\n"
               "		<hm:outbox/>\n"
               "		<hm:sendmsg/>\n"
               "		<hm:sentitems/>\n"
               "		<hm:deleteditems/>\n"
               "		<hm:drafts/>\n"
               "		<hm:msgfolderroot/>\n"
               "		<h:maxpoll/>\n"
               "		<h:sig/>\n"
               "	</D:prop>\n"
               "</D:propfind>\n", String::STATIC);

}


// External Global Variables ===============================================================================================

// Private Functions =======================================================================================================

namespace {

	Array<String> priv_stripEmailAddresses( String list )
	{
		StringTokenizer toker;
		toker.addContainer("\"", "\"", StringTokenizer::EXCLUSIVE);

		Array<String> addrs;
		Array<String> tokens = toker.tokenize(list, ", <>");
		for( int i = 0; i < tokens.getSize(); ++i ) {
			String token = tokens[i];
			if( token.beginsWith("\"") && token.endsWith("\"") ) {
				continue;
			}

			if( token.equalsWildCard("*@*.*") ) {
				addrs.append(token);
			}
		}

		return (addrs);
	}

	String priv_getHttpMailServerUrl( String domain )
	{
		if( domain.equalsIgnoreCase("hotmail.com") ) {
			return $("http://services.msn.com/svcs/hotmail/httpmail.asp");
		}
		else if( domain.equalsIgnoreCase("msn.com") ) {
			return $("http://oe.msn.msnmail.hotmail.com/cgi-bin/hmdata");
		}
		else if( domain.equalsIgnoreCase("lycos.co.uk") ) {
			return $("http://webdav.lycos.co.uk/httpmail.asp");
		}
		else if( domain.equalsIgnoreCase("lycos.ch") || domain.equalsIgnoreCase("lycos.de")) {
			return $("http://webdav.lycos.de/httpmail.asp");
		}
		else if( domain.equalsIgnoreCase("lycos.es") ) {
			return $("http://webdav.lycos.es/httpmail.asp");
		}
		else if( domain.equalsIgnoreCase("lycos.it") ) {
			return $("http://webdav.lycos.it/httpmail.asp");
		}
		else if( domain.equalsIgnoreCase("lycos.at") ) {
			return $("http://webdav.lycos.at/httpmail.asp");
		}
		else if( domain.equalsIgnoreCase("lycos.nl") ) {
			return $("http://webdav.lycos.nl/httpmail.asp");
		}
		else if( domain.equalsIgnoreCase("spray.se") ) {
			return $("http://webdav.spray.se/httpmail.asp");
		}
		else {
			return (String::null);
		}
	}

}

// Functions ===============================================================================================================

namespace blue {
namespace ext {
namespace httpmail {

	// ---------------------------------------------------------------------------------------------------------------------

	Mailbox::Mailbox() :m_session(0), m_root(0), m_httpLog(false)
	{
		m_loginUrl     = String::null;
		m_loginContent = g_defLoginContent;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	Mailbox::~Mailbox()
	{
		logout();
	}

	// ---------------------------------------------------------------------------------------------------------------------

	bool Mailbox::getHttpLogging() const
	{
		return (m_httpLog);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Mailbox::getHttpLoggingFile() const
	{
		return (m_httpLogFile);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Mailbox::getHttpProxyHost() const
	{
		return (m_proxyHost);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	uint16_t Mailbox::getHttpProxyPort() const
	{
		return (m_proxyPort);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	bool Mailbox::isLoggedIn() const
	{
		return (m_session != 0);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Mailbox::getLoginUrl() const
	{
		return (m_loginUrl);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Mailbox::getLoginContent() const
	{
		return (m_loginContent);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Mailbox::getUser() const
	{
		return (m_user);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Mailbox::getPassword() const
	{
		return (m_password);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Mailbox::getDomain() const
	{
		return (m_domain);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	Session* Mailbox::getSession()
	{
		return (m_session);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	Folder* Mailbox::getRootFolder()
	{
		return (m_root);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	const Folder* Mailbox::getRootFolder() const
	{
		return (m_root);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	const Session* Mailbox::getSession() const
	{
		return (m_session);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Mailbox::setHttpLogging( bool logging )
	{
		m_httpLog = logging;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Mailbox::setHttpLoggingFile( String file )
	{
		m_httpLogFile = file;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Mailbox::setHttpProxy( String host, uint16_t port )
	{
		m_proxyHost = host;
		m_proxyPort = port;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Mailbox::setLoginUrl( String url )
	{
		if( m_session != 0 ) {
			throw MailboxInvalidModeException("Cannot set login url while logged in");
		}
		m_loginUrl = url;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Mailbox::setLoginContent( String content )
	{
		if( m_session != 0 ) {
			throw MailboxInvalidModeException("Cannot set login content while logged in");
		}
		m_loginContent = content;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Mailbox::setUser( String user )
	{
		if( m_session != 0 ) {
			throw MailboxInvalidModeException("Cannot set user while logged in");
		}
		m_user = user;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Mailbox::setPassword( String password )
	{
		if( m_session != 0 ) {
			throw MailboxInvalidModeException("Cannot set password while logged in");
		}
		m_password = password;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Mailbox::login()
	{
		if( m_session != 0 ) {
			throw MailboxInvalidModeException("Cannot login while already logged in");
		}

		if( m_httpLog && m_httpLogFile != String::null ) {
			m_session = new LogSession(m_httpLogFile);
		}
		else {
			m_session = new Session();
		}

		try {
			if( m_loginUrl == String::null ) {
				m_domain = $("hotmail.com");
				int posDomain = m_user.findPos("@");
				if( posDomain != String::npos ) {
					m_domain = m_user.stripFromLeft(posDomain + 1);
					m_user   = m_user.left(posDomain);
				}
				m_loginUrl = priv_getHttpMailServerUrl(m_domain);
			}

			m_session->setFlags(Session::FOLLOW_REDIRECTED|Session::AUTHENTICATE);
			m_session->setPasswordRetriever( new HttpMailPasswordRetriever(m_user + "@" + m_domain, m_password) );
			m_session->setProxy(m_proxyHost, m_proxyPort);

			Url url(m_loginUrl);

			String errMsg;
			m_loginResponseContent = processWebDavRequest(url, "PROPFIND", m_loginContent, errMsg);

			if( m_loginResponseContent == String::null ) {
				throw MailboxLoginException("Unable to login to mailbox: " + errMsg);
			}

			data::BufferDevice device(m_loginResponseContent);
			data::InputStream parserStream(&device, true);
			xml::DomParser parser(&parserStream);
			parser.parse();

			xml::XPathExpression xpath(parser.getDocumentNode());
			
			String folderListUrl = xpath.findValue("//hm:msgfolderroot");
			m_root = new Folder(*this, 0, "root:" + folderListUrl);
			m_root->refresh();

			m_sendMessageUrl = xpath.findValue("//hm:sendmsg");
		}
		catch( ... ) {
			logout();
			throw;
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Mailbox::logout()
	{
		if( m_session == 0 ) {
			return;
		}

		if( m_root != 0 ) {
			delete m_root;
			m_root = 0;
		}

		if( m_session->getPasswordRetriever() != 0 ) {
			delete m_session->getPasswordRetriever();
			m_session->setPasswordRetriever(0);
		}

		delete m_session;
		m_session = 0;

		m_loginUrl             = String::null;
		m_loginContent         = g_defLoginContent;
		m_loginResponseContent = String::null;

		m_user     = String::null;
		m_password = String::null;
	}

	// ---------------------------------------------------------------------------------------------------------------------

	void Mailbox::sendMessage( String msg, String mailFromRcpt )
	{
		if( m_session == 0 ) {
			throw MailboxInvalidModeException("Mailbox must be logged in before email can be sent");
		}

		String invalidEmail = String("The email message to send must be a valid email", String::STATIC);

		String smtpCmd;
		if( mailFromRcpt == String::null ) {
			int posHeadEnd = msg.findPos("\r\n\r\n");
			if( posHeadEnd == String::npos ) {
				throw MailboxMessageSendException(invalidEmail);
			}
			String headers = msg.left(posHeadEnd);

			Array<String> fields = util::StringTokenizer().tokenize(headers, "\r\n");
			for( int i = 0; i < fields.getSize(); ++i ) {
				String field = fields[i];
				String prefix;

				int next = i+1;
				for( int iN = next; iN < fields.getSize(); ++iN ) {
					if( fields[iN].left(1).trim().getLength() == 0 ) {
						field += fields[iN];
					}
					else break;
				}

				if( field.beginsWithIgnoreCase("From:") ) {
					field = field.stripFromLeft(5);
					prefix = String("MAIL FROM:<", String::STATIC);
				}
				if( field.beginsWithIgnoreCase("To:") ||
					field.beginsWithIgnoreCase("Cc:") ||
					field.beginsWithIgnoreCase("Bcc:") ) {

					field = field.stripFromLeft( field.findPos(":") + 1 );
					prefix = String("RCPT TO:<", String::STATIC);
				}

				Array<String> addrs = priv_stripEmailAddresses(field);
				for( int i = 0; i < addrs.getSize(); ++i ) {
					smtpCmd += prefix + addrs[i] + String(">\r\n", String::STATIC);
				}
			}
		}
		else {
			smtpCmd = mailFromRcpt;
		}

		if( smtpCmd == String::null ) {
			throw MailboxMessageSendException(invalidEmail);
		}
		smtpCmd += String("\r\n", String::STATIC);

		String actlMessage = smtpCmd + msg;

		Headers custHeaders;
		custHeaders.setField(FIELD_CONTENT_TYPE, String("message/rfc821", String::STATIC));

		String errMsg;
		processWebDavRequest(Url(m_sendMessageUrl), String("POST", String::STATIC), actlMessage, custHeaders, errMsg);

		if( errMsg != String::null && !errMsg.equalsIgnoreCase("continue") ) {
			throw MailboxMessageSendException( String("The email message could not be sent", String::STATIC) );
		}
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Mailbox::processWebDavRequest( const Url& url, String webDavMethod, String webDavContent, String& errMsg )
	{
		Headers custHeaders;
		return processWebDavRequest(url, webDavMethod, webDavContent, custHeaders, errMsg);
	}

	// ---------------------------------------------------------------------------------------------------------------------

	String Mailbox::processWebDavRequest( const Url& url, String webDavMethod, String webDavContent, Headers& custHeaders, String& errMsg )
	{
		if( m_session == 0 ) {
			throw MailboxInvalidModeException("Mailbox must be logged in before processing any requests");
		}

		Request request;
		request.setMethod(webDavMethod);

		if( webDavContent != String::null ) {
			String contentType = String("text/xml", String::STATIC);
			if( custHeaders.hasField(FIELD_CONTENT_TYPE) ) {
				contentType = custHeaders.getField(FIELD_CONTENT_TYPE);
			}
			request.setContent(webDavContent, contentType);
		}

		request.accessHeaders().setField(FIELD_USER_AGENT, g_defUserAgent);

		Array<String> fields = custHeaders.getFields();
		for( int iF = 0; iF < fields.getSize(); ++iF ) {
			int count = custHeaders.getFieldCount(fields[iF]);
			for( int i = 0; i < count; ++i ) {
				request.accessHeaders().setField(fields[iF], i, custHeaders.getField(fields[iF],i));
			}
		}

		Response response;

		try {
			m_session->processRequest(url, request, response);
		}
		catch( SessionException& se ) {
			errMsg = se.toString();
			return (String::null);
		}

		if( !response.responseCodeMatches(Response::RC2xx) ) {
			errMsg = response.getResponseMessage();
			return (String::null);
		}

		String responseContent = response.getContent().toString();
		if( responseContent.findPosIgnoreCase("windows-1252") != String::npos ) {
			// The Xml  parser does not recognize this encoding scheme, but for all
			// intents and purposes, it's identical to UTF-8.
			responseContent = responseContent.replaceAllIgnoreCase("windows-1252", "utf-8");
		}

		return (responseContent);
	}

}}}	// namespaces
