//************************************************************************************************************************** 
//* Blue HttpMail Extension
//* Copyright (c) 2003-2004 Josh Harler
//* 
//* Blue - General Purpose C++ Library
//* Copyright (c) 2002-2004 Josh Harler
//* 
//* This software is provided 'as-is', without any express or implied warranty. In no event
//* will the authors be held liable for any damages arising from the use of this software.
//* 
//* Permission is granted to anyone to use this software for any purpose, including commercial
//* applications, and to alter it and redistribute it freely, subject to the following restrictions:
//* 
//* 	1. The origin of this software must not be misrepresented; you must not claim that you
//* 	wrote the original software. If you use this software in a product, an acknowledgment in the
//* 	product documentation would be appreciated but is not required.
//* 
//* 	2. Altered source versions must be plainly marked as such, and must not be misrepresented as
//* 	being the original software.
//* 
//* 	3. This notice may not be removed or altered from any source distribution.
//*
//*
//* file   Blue/Extension/HttpMail/Mailbox.h
//**

#ifndef __blue_ext_httpmail__file__h_included__
#define __blue_ext_httpmail__file__h_included__

// Public Headers ==========================================================================================================

// Extension headers
#include "Blue/Extension/HttpMail/Folder.h"

// Http Extension headers
#include "Blue/Extension/Http/Session.h"


// Public Defines/Enums/Typedefs/Etc. ======================================================================================

// Public Classes/Structs ==================================================================================================

namespace blue {
namespace ext {
namespace httpmail {

	/**
	 * \class Mailbox
	 * \brief Logs into the HttpMail account.
	 * \ingroup HttpMail
	 *
	 * 
	 */
	class BLUE_EXPORT_HTTPMAIL Mailbox
	{
	public:
		// ===========================================================
		//  creation/destruction
		// ===========================================================

		/** Constructor. */
		Mailbox();

		/** Destructor. */
		~Mailbox();


		// ===========================================================
		//  query
		// ===========================================================

		/** Determines if the HTTP communication will be logged. */
		bool getHttpLogging() const;
		/** If logging is set, this is the file it will log to. */
		String getHttpLoggingFile() const;

		/** Returns the Proxy host. */
		String getHttpProxyHost() const;
		/** Returns the Proxy port. */
		uint16_t getHttpProxyPort() const;

		/** Determines if the Mailbox is currently logged in. */
		bool isLoggedIn() const;

		/**
		 * Gets the login url.  This is the url used to connect to the
		 * HttpMail server.
		 */
		String getLoginUrl() const;
		/**
		 * Gets the WebDAV Xml content that is sent to the login url
		 * in order to obtain the information regarding the mailbox.
		 */
		String getLoginContent() const;

		/** Gets the user name of the mailbox account. */
		String getUser() const;
		/** Gets the password of the mailbox account. */
		String getPassword() const;
		/** Gets the domain of the mailbox account. */
		String getDomain() const;

		/**
		 * Gets the Http Session used for communicating with the
		 * server.  This function will return 0 when the mailbox is
		 * not logged in.
		 */
		http::Session* getSession();
		/** overload */
		const http::Session* getSession() const;

		/**
		 * Gets the root level folder of the mailbox.  When logging
		 * in, this folder has its Folder::refresh() function called,
		 * so it's not required to be called initially.  This function
		 * will return 0 when the mailbox is not logged in.
		 */
		Folder* getRootFolder();
		/** overload */
		const Folder* getRootFolder() const;


		// ===========================================================
		//  manipulation
		// ===========================================================

		/**
		 * Sets if the HTTP communication should be logged.  This
		 * must only be called before the mailbox logs in.
		 */
		void setHttpLogging( bool logging );
		/** If logging is set, this is the file it will log to.*/
		void setHttpLoggingFile( String file );

		/**
		 * Sets Proxy information if an HTTP proxy is to be used for
		 * communication.
		 */
		void setHttpProxy( String host, uint16_t port );

		/**
		 * Sets the login url.  This is the url used to connect to the
		 * HttpMail server.  If this function is called while the
		 * mailbox is logged in, a MailboxInvalidModeException is thrown.
		 */
		void setLoginUrl( String url );
		/**
		 * Sets the WebDAV Xml content that is sent to the login url
		 * in order to obtain the information regarding the mailbox.
		 * There is a default value set for this by the mailbox and
		 * it should not be changed unless you know what you are doing.
		 * If this function is called while the mailbox is logged in, a
		 * MailboxInvalidModeException is thrown.
		 */
		void setLoginContent( String content );

		/**
		 * Sets the user name of the mailbox account.  If this function is
		 * called while the mailbox is logged in, a MailboxInvalidModeException
		 * is thrown.
		 */
		void setUser( String user );
		/**
		 * Sets the password of the mailbox account.  If this function is
		 * called while the mailbox is logged in, a MailboxInvalidModeException
		 * is thrown.
		 */
		void setPassword( String password );

		/**
		 * Logs into the account.  The user name and password must have
		 * previously been set.  If this function is called while the mailbox
		 * is logged in, a MailboxInvalidModeException is thrown.
		 */
		void login();
		/**
		 * Logs out of the account.  Any Folder instances and Message instance
		 * are deleted and should be discarded.  They will not become valid again
		 * if the mailbox logs in again.
		 */
		void logout();

		/**
		 * Sends an email.  The <code>message</code> string must
		 * contain a valid email as specified in RFC 2822.
		 */
		void sendMessage( String msg, String mailFromRcpt = String::null );

		/**
		 * Processes a WebDAV request.
		 * \param url - The url to send the request to
		 * \param webDavMethod - The method of the request ("GET", "PROPFIND", etc)
		 * \param webDavContent - The Xml content to be sent along with the request
		 * \param errMsg - The String that will hold the error message if the request
		 *                 fails.
		 */
		String processWebDavRequest( const http::Url& url, String webDavMethod, String webDavContent, String& errMsg );
		/**
		 * Overloaded to allow for custom headers to be added to a request.
		 */
		String processWebDavRequest( const http::Url& url, String webDavMethod, String webDavContent, http::Headers& custHeaders, String& errMsg );


	private:
		/**
		 * Copy constructor. Private because Mailboxes should not be
		 * manipulated by more than one instance.
		 */
		Mailbox( const Mailbox& );
		/** Private assignment operator. See copy constructor documentation. */
		const Mailbox& operator=( const Mailbox& );

		bool   m_httpLog;
		String m_httpLogFile;

		String m_loginUrl;
		String m_loginContent;
		String m_loginResponseContent;
		String m_sendMessageUrl;

		String m_user;
		String m_password;
		String m_domain;

		String   m_proxyHost;
		uint16_t m_proxyPort;

		http::Session* m_session;
		http::Session::PasswordRetriever* m_passwordRetriever;

		Folder* m_root;
	};


	/**
	 * \class MailboxException
	 * \brief Base class for Mailbox exceptions.
	 * \ingroup HttpMailExceptions
	 */
	class BLUE_EXPORT_HTTPMAIL MailboxException :public HttpMailException
	{
	public:
		MailboxException( String desc )
			:HttpMailException(desc) {}

		virtual String getException() const {
			return ("MailboxException");
		}
	};


	/**
	 * \class MailboxLoginException
	 * \brief Thrown when the Mailbox fails to login successfully.
	 * \ingroup HttpMailExceptions
	 */
	class BLUE_EXPORT_HTTPMAIL MailboxLoginException :public MailboxException
	{
	public:
		MailboxLoginException( String desc )
			:MailboxException(desc) {}

		virtual String getException() const {
			return ("MailboxLoginException");
		}
	};

	
	/**
	 * \class MailboxInvalidModeException
	 * \brief Thrown when an operation is attempted on a Mailbox that's in
	 *        a state where the operation is not valid (setting user or password
	 *        while being logged in).
	 * \ingroup HttpMailExceptions
	 */
	class BLUE_EXPORT_HTTPMAIL MailboxInvalidModeException :public MailboxException
	{
	public:
		MailboxInvalidModeException( String desc )
			:MailboxException(desc) {}

		virtual String getException() const {
			return ("MailboxInvalidModeException");
		}
	};

	
	/**
	 * \class MailboxMessageSendException
	 * \brief Thrown when an attempt to send an email message from the Mailbox
	 *        failed.
	 * \ingroup HttpMailExceptions
	 */
	class BLUE_EXPORT_HTTPMAIL MailboxMessageSendException :public MailboxException
	{
	public:
		MailboxMessageSendException( String desc )
			:MailboxException(desc) {}

		virtual String getException() const {
			return ("MailboxMessageSendException");
		}
	};

}}}	// namespaces


// Public External Variables ===============================================================================================

// Public Function Prototypes ==============================================================================================

// Public Inline Functions =================================================================================================

#endif // include guard
