//************************************************************************************************************************** 
//* Blue HttpMail Extension
//* Copyright (c) 2003-2004 Josh Harler
//* 
//* Blue - General Purpose C++ Library
//* Copyright (c) 2002-2004 Josh Harler
//* 
//* This software is provided 'as-is', without any express or implied warranty. In no event
//* will the authors be held liable for any damages arising from the use of this software.
//* 
//* Permission is granted to anyone to use this software for any purpose, including commercial
//* applications, and to alter it and redistribute it freely, subject to the following restrictions:
//* 
//* 	1. The origin of this software must not be misrepresented; you must not claim that you
//* 	wrote the original software. If you use this software in a product, an acknowledgment in the
//* 	product documentation would be appreciated but is not required.
//* 
//* 	2. Altered source versions must be plainly marked as such, and must not be misrepresented as
//* 	being the original software.
//* 
//* 	3. This notice may not be removed or altered from any source distribution.
//*
//*
//* file   Blue/Extension/HttpMail/Message.h
//**

#ifndef __blue_ext_httpmail_Message_h_included__
#define __blue_ext_httpmail_Message_h_included__

// Public Headers ==========================================================================================================

// Extension headers
#include "Blue/Extension/HttpMail/HttpMail.h"

// Blue library headers
#include "Blue/Util/DateTime.h"

// Http Extension headers
#include "Blue/Extension/Http/Session.h"


// Public Defines/Enums/Typedefs/Etc. ======================================================================================

// Public Classes/Structs ==================================================================================================

namespace blue {
namespace ext {
namespace httpmail {

	/****
	 * Forward Declaration
	 */
	class Mailbox;


	/**
	 * \class Message
	 * \brief Allows retrieval of the message body.
	 * \ingroup HttpMail
	 *
	 * This class offers a way for applications to retrieve the
	 * content of an email message.
	 *
	 * Note that no manipulation to the actual message on the
	 * server is done through this class.  To manipulate the
	 * messages, use the Folder class.
	 */
	class BLUE_EXPORT_HTTPMAIL Message
	{
	public:
		// ===========================================================
		//  creation/destruction
		// ===========================================================

		/** Constructor. */
		Message( Mailbox& mailbox, String url );
		/** Destructor. */
		~Message();


		// ===========================================================
		//  query
		// ===========================================================

		/** Gets the url used to access this message. */
		String getUrl() const;

		/**
		 * Gets the ID of the message.  Each mail message has its
		 * own unique ID given to it by the server.
		 */
		String getID() const;

		/** Determines if the content of the message has been loaded. */
		bool isLoaded() const;
		/** Determines if the message is marked as read on the server. */
		bool isRead() const;

		/** Gets the length of the message in bytes. */
		int getContentLength() const;
		/**
		 * Gets the name or email address of the entity that sent this
		 * message.  If a textual name is provided, that is returned,
		 * otherwise, the email address is returned.
		 */
		String getFrom() const;
		/** Gets the subject of the message. */
		String getSubject() const;
		/** Gets the date of the message (the date the server received it). */
		util::DateTime getDate() const;

		/**
		 * Gets the content of the message.  The message must have
		 * been previously loaded in order for this to return valid
		 * data.
		 */
		const Buffer getContent() const;


		// ===========================================================
		//  manipulation
		// ===========================================================

		/** Gets the url used to access this message. */
		void setUrl( String url );

		/**
		 * Sets if the message is marked as read on the server.  This
		 * does NOT change the actual status of the message on the
		 * server.  To do that, call Folder::setMessageRead() on the
		 * folder where this message resides.
		 */
		void setRead( bool read );
		/**
		 * Sets the content length of the message.  This only affects
		 * the value returned by getContentLength().
		 */
		void setContentLength( int length );
		/**
		 * Sets the from entity of the message.  This only affects
		 * the value returned by getFrom().
		 */
		void setFrom( String from );
		/**
		 * Sets the subject of the message.  This only affects
		 * the value returned by getSubject().
		 */
		void setSubject( String subject );
		/**
		 * Sets the received date of the message.  This only affects
		 * the value returned by getDate().
		 */
		void setDate( util::DateTime date );

		/** Loads the message content into memory. */
		void load();
		/** Unloads the message content from memory. */
		void unload();


	private:
		/**
		 * Copy constructor. Private because Messages should not be
		 * manipulated by more than one instance.
		 */
		Message( const Message& );
		/** Private assignment operator. See copy constructor documentation. */
		const Message& operator=( const Message& );

		Mailbox&       m_mailbox;
		String         m_url;
		bool           m_isRead;
		Buffer         m_content;
		int            m_contentLength;
		String         m_from;
		String         m_subject;
		util::DateTime m_date;
	};



	/**
	 * \class MessageLoadException
	 * \brief Thrown when the message cannot be retrieved from the server.
	 * \ingroup HttpMailExceptions
	 */
	class BLUE_EXPORT_HTTPMAIL MessageLoadException :public HttpMailException
	{
	public:
		MessageLoadException( String desc )
			:HttpMailException(desc) {}

		virtual String getException() const {
			return ("MessageLoadException");
		}
	};

}}}	// namespaces


// Public External Variables ===============================================================================================

// Public Function Prototypes ==============================================================================================

// Public Inline Functions =================================================================================================

#endif // include guard
